#include "../StdAfx.h"
#include "DTV_EzUsbfx2.h"

#pragma comment(lib,"cyapi.lib")

CDTV_EzUSBFX2::CDTV_EzUSBFX2() : CDTV_Base()
{
	m_pCyUSB=NULL;
	m_bHvMikittyInit=FALSE;
	CDTV_CriticalSection::InitializeCriticalSection(&m_stCS);
}

CDTV_EzUSBFX2::~CDTV_EzUSBFX2(void)
{
	if(IsInitialized()) Release();
	CDTV_CriticalSection::DeleteCriticalSection(&m_stCS);
}

BOOL CDTV_EzUSBFX2::Initialize(char *szInitString,HWND hWnd,GUID *pGuid)
{
	if(IsInitialized()) return FALSE;
	if(!pGuid) pGuid=&CYUSBDRV_GUID;
	m_pCyUSB=new CCyUSBDevice(hWnd,*pGuid);
	if(!m_pCyUSB) return FALSE;
	return TRUE;
}

void CDTV_EzUSBFX2::Release()
{
	DisConnect();
	if(m_pCyUSB) delete m_pCyUSB;
	m_pCyUSB=NULL;
}

BOOL CDTV_EzUSBFX2::IsInitialized()
{
	return m_pCyUSB?TRUE:FALSE;
}

DWORD CDTV_EzUSBFX2::GetExtFunction()
{
	return DTV_EXTFUNC_NONE;
}

DWORD CDTV_EzUSBFX2::GetDeviceCount()
{
	if(!IsInitialized()) return 0;
	return m_pCyUSB->DeviceCount();
}

BOOL CDTV_EzUSBFX2::GetDeviceName(int iDevNo,char *lpszDevName,int iBufLen)
{
	if(!IsInitialized()) return FALSE;
	if(m_pCyUSB->Open(iDevNo))
	{
		lstrcpyn(lpszDevName,m_pCyUSB->DeviceName,iBufLen);
		if(!lpszDevName[0])
		{
			lstrcpyn(lpszDevName,"?",iBufLen);
		}
		m_pCyUSB->Close();
		return TRUE;
	}
	return FALSE;
}

DWORD CDTV_EzUSBFX2::GetUsbEndpointCount(int iDevNo)
{
	if(!IsInitialized()) return 0;
	if(m_pCyUSB->Open(iDevNo))
	{
		UCHAR iEpCount=m_pCyUSB->EndPointCount();
		m_pCyUSB->Close();
		return iEpCount;
	}
	return 0;
}

BOOL CDTV_EzUSBFX2::GetUsbEndpointInfo(int iDevNo,int iUsbEpNo,struct DTV_UsbEndpointInfo *pUsbEpInfo)
{
	if(!IsInitialized()) return FALSE;
	if(m_pCyUSB->Open(iDevNo))
	{
		BOOL bRet=FALSE;
		CCyUSBEndPoint *pCyEP=m_pCyUSB->EndPoints[iUsbEpNo];
		if(pCyEP)
		{
			if(pUsbEpInfo)
			{
				pUsbEpInfo->DscLen=pCyEP->DscLen;
				pUsbEpInfo->DscType=pCyEP->DscType;
				pUsbEpInfo->Address=pCyEP->Address;
				pUsbEpInfo->Attributes=pCyEP->Attributes;
				pUsbEpInfo->MaxPktSize=pCyEP->MaxPktSize;
				pUsbEpInfo->PktsPerFrame=pCyEP->PktsPerFrame;
				pUsbEpInfo->Interval=pCyEP->Interval;
				pUsbEpInfo->TimeOut=pCyEP->TimeOut;
				pUsbEpInfo->UsbdStatus=pCyEP->UsbdStatus;
				pUsbEpInfo->NtStatus=pCyEP->NtStatus;				
			}
			bRet=TRUE;
		}
		m_pCyUSB->Close();
		return bRet;
	}
	return FALSE;
}

BOOL CDTV_EzUSBFX2::IsEndpointEffective(int iDevNo,int iUsbEpAddr)
{
	if(!IsInitialized()) return FALSE;
	BOOL bRet=FALSE;
	int i,n=GetUsbEndpointCount(iDevNo);	
	for(i=0;i<n&&!bRet;i++)
	{
		struct DTV_UsbEndpointInfo stUsbEpInfo;
		if(GetUsbEndpointInfo(iDevNo,i,&stUsbEpInfo))
		{
			if(stUsbEpInfo.Address==iUsbEpAddr) bRet=TRUE;
		}
	}
	return bRet;
}

BOOL CDTV_EzUSBFX2::Connect(int iDevNo,int iInEpAddr,int iOutEpAddr)
{
	if(!IsInitialized()) return FALSE;
	if(IsConnect()) return FALSE;

	struct DTV_ThreadParam *pReq=NULL;
	struct DTV_ThreadParam *pRes=NULL;
	try
	{
		pReq=new struct DTV_ThreadParam;
		pRes=new struct DTV_ThreadParam;
		if(!pReq||!pRes) throw 101;
		ZeroMemory(pReq,sizeof(struct DTV_ThreadParam));
		ZeroMemory(pRes,sizeof(struct DTV_ThreadParam));
		pReq->pClassPtr=pRes->pClassPtr=this;
		m_stThread.hExitEvent=pReq->hExitEvent=pRes->hExitEvent=CreateEvent(NULL,TRUE,FALSE,NULL);
		if(!m_stThread.hExitEvent) throw 102;

		struct DTV_EzUSBFX2_ReqParam *pReqParam=new struct DTV_EzUSBFX2_ReqParam;
		if(!pReqParam) throw 103;
		ZeroMemory(pReqParam,sizeof(struct DTV_EzUSBFX2_ReqParam));
		pReq->pData=pReqParam;

		// Gh|Cg
		int iInPos,iOutPos;
		int nEP=GetUsbEndpointCount(iDevNo);
		iInPos=iOutPos=-1;
		for(int i=0;i<nEP;i++)
		{
			struct DTV_UsbEndpointInfo stUsbEP;
			if(GetUsbEndpointInfo(iDevNo,i,&stUsbEP))
			{
				if(iInEpAddr==stUsbEP.Address) iInPos=i;
				if(iOutEpAddr==stUsbEP.Address) iOutPos=i;
			}
		}
		// foCXJăGh|Cg擾
		if(!m_pCyUSB->Open(iDevNo)) throw 301;
		if(iInPos!=-1) pReqParam->pEP_In=m_pCyUSB->EndPoints[iInPos];
		if(iOutPos!=-1) pReqParam->pEP_Out=m_pCyUSB->EndPoints[iOutPos];

		// 񓯊L[̍쐬
		if(!MakeXferQueue())
		{
			throw 401;
		}

		// ]󋵂̃Zbg
		ClearTransferInfo();
		
		m_stThread.stRequest.hThread=CreateThread(NULL,0,XferRequestThreadEntry,pReq,CREATE_SUSPENDED,&m_stThread.stRequest.dwThreadID);
		m_stThread.stResponce.hThread=CreateThread(NULL,0,XferResponceThreadEntry,pRes,CREATE_SUSPENDED,&m_stThread.stResponce.dwThreadID);
		if(!m_stThread.stRequest.hThread||!m_stThread.stResponce.hThread)
		{
			throw 501;
		}

		// Jn
		ResumeThread(m_stThread.stRequest.hThread);
		ResumeThread(m_stThread.stResponce.hThread);
	}
	catch(int err_code)
	{
		DTV_CATCH_INT_ERROR_CODE(err_code);
		if(m_stThread.stRequest.hThread)
			TerminateThread(m_stThread.stRequest.hThread,0xffffffff);
		if(m_stThread.stResponce.hThread)
			TerminateThread(m_stThread.stResponce.hThread,0xffffffff);
		if(m_stThread.hExitEvent)
		{
			CloseHandle(m_stThread.hExitEvent);
		}
		ZeroMemory(&m_stThread,sizeof(struct DTV_CurrentThread));

		FreeXferQueue();
		if(pReq)
		{
			if(pReq->pData) delete pReq->pData;
			delete pReq;
		}
		if(pRes)
		{
			if(pRes->pData) delete pRes->pData;
			delete pRes;
		}
		if(IsConnect())
		{
			m_pCyUSB->Close();
		}
		return FALSE;
	}
	return TRUE;
}

void CDTV_EzUSBFX2::DisConnect()
{
	if(IsConnect())
	{
		SetEvent(m_stThread.hExitEvent);
		HANDLE hWait[2]={m_stThread.stRequest.hThread,m_stThread.stResponce.hThread};
		if(WaitForMultipleObjects(2,hWait,TRUE,10*1000)!=WAIT_OBJECT_0)
		{
			// ^CAEgȂ̂ŋ~
			TerminateThread(m_stThread.stRequest.hThread,0xffffffff);
			TerminateThread(m_stThread.stResponce.hThread,0xffffffff);
		}
		DWORD dwExitCode;
		GetExitCodeThread(m_stThread.stRequest.hThread,&dwExitCode);
		GetExitCodeThread(m_stThread.stResponce.hThread,&dwExitCode);

		// foCX
		m_pCyUSB->Close();

		// L[J
		FreeXferQueue();

		// nh
		CloseHandle(m_stThread.stRequest.hThread);
		CloseHandle(m_stThread.stResponce.hThread);
		CloseHandle(m_stThread.hExitEvent);
		ZeroMemory(&m_stThread,sizeof(struct DTV_CurrentThread));
	}
}

BOOL CDTV_EzUSBFX2::IsConnect()
{
	if(!IsInitialized()) return FALSE;
	return m_pCyUSB->IsOpen();
}

DWORD CDTV_EzUSBFX2::XferRequestProc(struct DTV_ThreadParam *lpParameter)
{
	struct DTV_EzUSBFX2_ReqParam *pReqParam=(struct DTV_EzUSBFX2_ReqParam*)lpParameter->pData;
	bool bRet;
	LONG iRecvOut;

	// Hv-Mikitty t@[p
	if(m_bHvMikittyInit&&pReqParam->pEP_Out)
	{
		BYTE bufOut[10];
		// Send 0x8000
		iRecvOut=2;
		bufOut[0]=0x80;
		bufOut[1]=0x00;
		bRet=pReqParam->pEP_Out->XferData(bufOut,iRecvOut);
		// Send 0x8B
		iRecvOut=1;
		bufOut[0]=0x8B;
		bRet=pReqParam->pEP_Out->XferData(bufOut,iRecvOut);
		// Send 0x8001
		iRecvOut=2;
		bufOut[0]=0x80;
		bufOut[1]=0x01;
		bRet=pReqParam->pEP_Out->XferData(bufOut,iRecvOut);
	}

	// ]f[^TCY
	pReqParam->pEP_In->SetXferSize(DTV_EzUSBFX2_ASYNC_XFER_SIZE);

    // NGXg[v
	while(WaitForSingleObject(lpParameter->hExitEvent,0)!=WAIT_OBJECT_0)
	{
		struct DTV_EzUSBFX2_Request *pItem=NULL;			
		if(m_ReqWait.size()>0)
		{
			// Wait쏜
			{
				CDTV_CriticalSection sec(&m_stCS);
				pItem=m_ReqWait.front();
				m_ReqWait.pop_front();
			}
			ZeroMemory(&pItem->Overlap,sizeof(OVERLAPPED));
			pItem->Overlap.hEvent=CreateEvent(NULL,FALSE,FALSE,NULL);
			pItem->pEndPoint=pReqParam->pEP_In;
			pItem->buf_len=DTV_EzUSBFX2_ASYNC_XFER_SIZE;
#ifdef _DEBUG
			ZeroMemory(pItem->buf,pItem->buf_len);
#endif
			pItem->pContext=pItem->pEndPoint->BeginDataXfer(pItem->buf,pItem->buf_len,&pItem->Overlap);
			// Busyɒǉ
			{
				CDTV_CriticalSection sec(&m_stCS);
				m_ReqBusy.push_back(pItem);
			}
		}		
		Sleep(1);
	}

	// I
	if(pReqParam)
	{
		if(pReqParam->pEP_In) pReqParam->pEP_In->Abort();
		if(pReqParam->pEP_Out) pReqParam->pEP_Out->Abort();
		delete pReqParam;
	}
	delete lpParameter;
	ExitThread(1);
	return 0;
}

DWORD CDTV_EzUSBFX2::XferResponceProc(struct DTV_ThreadParam *lpParameter)
{

    // M[v
	while(WaitForSingleObject(lpParameter->hExitEvent,0)!=WAIT_OBJECT_0 || m_ReqBusy.size()>0)
	{
		struct DTV_EzUSBFX2_Request *pItem=NULL;
		bool bRet;			
		if(m_ReqBusy.size()>0)
		{
			// Busy̐擪擾
			{
				CDTV_CriticalSection sec(&m_stCS);
				pItem=m_ReqBusy.front();
				m_ReqBusy.pop_front();
			}

			// ̏ bRet=WaitForSingleObject(pItem->Overlap.hEvent,1000)==WAIT_OBJECT_0?true:false; Ɠۂ
			bRet=pItem->pEndPoint->WaitForXfer(&pItem->Overlap,1000);
			if(bRet)			
			{
				// ł̓GetOverlappedResult(bWait==FALSE)
				bRet=pItem->pEndPoint->FinishDataXfer(pItem->buf,pItem->buf_len,&pItem->Overlap,pItem->pContext);
				if(bRet)
				{
					if(pItem->buf_len>0)
					{
						m_stTransferInfo.ui64RecvTotal+=pItem->buf_len;
						DoDTVDataCallback(pItem->buf,pItem->buf_len,m_stTransferInfo.ui64RecvTotal);						
					}
					m_stTransferInfo.dwSuccess++;
				} else {
					m_stTransferInfo.dwFail++;
				}
			} else {
				m_stTransferInfo.dwFail++;
			}
			CloseHandle(pItem->Overlap.hEvent);
			pItem->Overlap.hEvent=NULL;
			// Waitɒǉ
			{
				CDTV_CriticalSection sec(&m_stCS);
				m_ReqWait.push_back(pItem);
			}
		}			

		Sleep(1);
	}

	// I
	delete lpParameter;
	ExitThread(1);
	return 0;
}

// 񓯊L[쐬
BOOL CDTV_EzUSBFX2::MakeXferQueue()
{
	for(int i=0;i<DTV_EzUSBFX2_ASYNC_XFER_REQ_NUM;i++)
	{
		struct DTV_EzUSBFX2_Request *pItem=new struct DTV_EzUSBFX2_Request;
		if(!pItem)
		{
			FreeXferQueue();
			return FALSE;
		}
		ZeroMemory(pItem,sizeof(struct DTV_EzUSBFX2_Request));
		m_ReqWait.push_back(pItem);
	}
	return TRUE;
}

// 񓯊L[j
void CDTV_EzUSBFX2::FreeXferQueue()
{
	struct DTV_EzUSBFX2_Request *pItem;
	// ҋ@
	int iWait=0,iBusy=0;
	while(m_ReqWait.begin()!=m_ReqWait.end())
	{
		pItem=m_ReqWait.front();
		delete pItem;
		m_ReqWait.pop_front();
		iWait++;
	}
	// NGXg
	while(m_ReqBusy.begin()!=m_ReqBusy.end())
	{
		pItem=m_ReqBusy.front();
		bool bRet=pItem->pEndPoint->WaitForXfer(&pItem->Overlap,1000);
		if(pItem->Overlap.hEvent) CloseHandle(pItem->Overlap.hEvent);
		delete pItem;
		m_ReqBusy.pop_front();
		iBusy++;
	}
}
